/*
randomized random-cipher encryption version 2
rrce License 200708231502

Copyright (c) 2007 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end rrce License



randomized random-cipher encryption
encrypts 1 byte at a time by xor with 1 byte key segment from key file
key segment pointer is determined from nonce/iv, input data, and key file data
randomized order of key segments is determined by nonce/iv, key file data, and input data

any change in input data, byte number, or iv will result in a different key value used

this encryption algorithm aims for greater speed and simplicity than RRCEV1

8-bit algorithm, so key can range from 16 to 256 bytes (128 to 2048 bits)

--byte / algorithm layout for 3-byte file--
nonce / initialization vector iv (previousEncryptedByte = 0)
input[0] = input[0] ^ key[key[nonceiv+0]] ^ key[previousEncryptedByte]
input[1] = input[1] ^ key[key[nonceiv+1]] ^ key[previousEncryptedByte]
input[2] = input[2] ^ key[key[nonceiv+2]] ^ key[previousEncryptedByte]
--end byte layout--

The output filesize = inputfilesize+1.

The keys can be 256bytes, but it revolves
for smaller key files. This makes it very difficult to determine key size.

The entire key is also not guaranteed to be used on any one file.

---Known Plaintext Attack, which already constitutes a security breach---
In the case of a known plaintext attack, segments of the key can be easily
determined by xoring the ciphertext and known plaintext. However, the order and
extent of key segments is unknown, so the original key cannot be easily reconstructed.

In the case that a set of key segments is determined from a known plaintext
attack and another encrypted file exists with the same initial key offset,
additional data are kept safe by the computation of offsets from unknown
key data xored with ciphertext data. This ensures that neither the same
key segments nor the same order are used on different files starting with the same offset.

Known plaintext attack works iff (if and only if) the entire key was used, key size
is known, and all segments were compromised; otherwise permutations increase.
Assume a very small key size of 64 Bytes (512bits) for this example.
Key segment size is 1byte (8bits), but segment can start at any byte in key.
Number of key segments possible: 64.
All possible keys from different segment ordering: 64! ~= 1.27x10^89

Now imagine a key that is 100B (800bits).
Key segment size is still 1byte (8bits), but segment can start at any byte in key.
Now try to determine the correct key based on the order of the 100 segments
All possible keys from different segment ordering: 100! ~= 9.33x10^157

Now imagine a 1MB or 1GB key.

---Known Plaintext Attack, which already constitutes a security breach---


There may very well be a shortcut method to breaking this encryption, but I have not found it yet.


*/


#ifndef RRCE_H
#define RRCE_H

#include <string>
#include <fstream>
#include <time.h>
using namespace std;

class rrce{
public:
	rrce();//default constructor
	~rrce();//default destructor
	void loadkey(char*);//loads secret encryption key, must be over 128bits or 16bytes
	void encrypt(unsigned char*,unsigned long long,unsigned char);//encrypts a buffer, assuming key loaded and initial offset computed
	void decrypt(unsigned char*,unsigned long long,unsigned char);//decrypts a buffer, assuming key loaded and initial offset read
private:
	unsigned char* key;
	unsigned char keysize;
	unsigned char byte;
	unsigned char previousbyte;
	unsigned char power;
	bool debug;
};//end class rrce

rrce::rrce(){
	debug = false;;
	key = NULL;
	keysize = 0;
	byte = 0;
	previousbyte = 0;
}//constructor

rrce::~rrce(){
	for(unsigned char i=0; i < keysize; i++){
		key[i] = 0x00;
	}//end for
	free(key);
	keysize = 0;
	byte = 0;
	previousbyte = 0;
}//destructor

//load secret encryption key
void rrce::loadkey(char* keyfilename){
	ifstream keyfile;
	keyfile.open(keyfilename,ios::binary);
	if(!keyfile.good()){
		cerr << "Error: could not load key file " << keyfilename << endl;
		exit(1);
	}//end if
	keyfile.seekg(0,ios::end);
	keysize = keyfile.tellg();
	keyfile.seekg(0,ios::beg);
	key = NULL;
	key = (unsigned char*)malloc(keysize);
	if(key == NULL){
		cerr << "Error: key malloc failed for keysize " << keysize << endl;
		exit(2);
	}//end if
	keyfile.read((char*)key,keysize);
	keyfile.close();
	if(keysize < 16){
		cerr << "Error: keysize is too small!\n";
		exit(4);
	}//end if
	unsigned char testval = 0;
	power = 0;
	while(testval < keysize){
		power++;
		testval = 1<<power;
	}
	cout << "keysize is 2^" << (unsigned short)power << endl;
	if(keysize != testval){
		cout << "keysize: " << (unsigned short)keysize << " power: " << (unsigned short)power << endl;
		cerr << "Error: keysize not a power of 2!\n";
		exit(4);
	}
}//end loadkey

//encrypt
void rrce::encrypt(unsigned char* buffer, unsigned long long buffersize, unsigned char nonceiv){
	unsigned long long s,d;
	unsigned char r,k1,k2,k3;
	previousbyte = 0;
	for(unsigned long long i=0; i < buffersize; i++){
		//read input byte
		byte = buffer[i];

		//xor encrypt byte using input data, key data, and nonceiv
		////short slow formula -- any key size
		//byte ^= key[key[(i+nonceiv)%keysize]%keysize] ^ key[previousbyte%keysize];
		////long fast formula -- power of 2 key size
		s= i+nonceiv;
		d = s>>power;
		d <<= power;
		r = s-d;
		k1 = key[r];
		d = k1>>power;
		d <<= power;
		r = k1-d;
		k2 = key[r];
		d = previousbyte>>power;
		d <<= power;
		r = previousbyte-d;
		k3 = key[r];
		byte ^= k2 ^ k3;

		//write output byte
		previousbyte = byte;
		buffer[i] = byte;
	}//end for
}//end encrypt

//decrypt
void rrce::decrypt(unsigned char* buffer, unsigned long long buffersize,unsigned char nonceiv){
	unsigned long long s,d;
	unsigned char r,k1,k2,k3;
	previousbyte = 0;
	for(unsigned long long i=0; i < buffersize; i++){
		//read input byte
		byte = buffer[i];

		//xor encrypt byte using input data, key data, and nonceiv
		////short slow formula -- any key size
		//byte ^= key[key[(i+nonceiv)%keysize]%keysize] ^ key[previousbyte%keysize];
		////long fast formula -- power of 2 key size
		s= i+nonceiv;
		d = s>>power;
		d <<= power;
		r = s-d;
		k1 = key[r];
		d = k1>>power;
		d <<= power;
		r = k1-d;
		k2 = key[r];
		d = previousbyte>>power;
		d <<= power;
		r = previousbyte-d;
		k3 = key[r];
		byte ^= k2 ^ k3;

		//write output byte
		previousbyte = buffer[i];
		buffer[i] = byte;
	}//end for
}//end decrypt


#endif

